/**
 *
 * Centralina mobile (ESP32 + KS0578 + LED RGB + DHT22)
 *
 *   GUIDA PASSO PASSO per costruire la tua CENTRALINA FAI DA TE
 *   👉️ https://comunitacircolare.it/articoli/scienziaria-guida-centralina
 *
 * Questa variante permette un monitoraggio continuo ed è ispirata al codice
 * della centralina mobile del progetto ScienziAria di Comunità Circolare.
 *
 * Oltre alle polveri sottili, rileva anche temperatura e umidità relativa. Può
 * essere alimentata con un power bank per smartphone e connettersi via hotspot
 * Wi-Fi per aggiornare periodicamente i grafici sul canale ThingSpeak.
 *
 * Facendo riferimento alle schede tecniche dei componenti, abbiamo stimato un
 * consumo di corrente medio non superiore ai 160 mA (che può comunque variare
 * a seconda delle effettive modalità di realizzazione e uso della centralina).
 *
 * In base alle impostazioni predefinite, la concentrazione di polveri sottili
 * viene misurata ogni secondo e segnalata da una spia luminosa. La centralina
 * carica dati online ogni 16 secondi circa, a meno di problemi di connettività.
 * Se connessa al Wi-Fi, la luce è fissa; altrimenti, l'indicatore lampeggia.
 *
 * COLORE LED  STATO DELLA CENTRALINA
 * Azzurro ... Stabilizzazione sensore PM
 * Verde ..... Concentrazione PM10 bassa
 * Blu ....... Concentrazione PM10 media
 * Rosso ..... Concentrazione PM10 alta
 * Bianco .... Misurazione non riuscita
 * Magenta ... Aggiornamento ThingSpeak
 *
 * CAMPI DEL CANALE THINGSPEAK
 * 1. PM10: media mobile esponenziale di rilevazioni dall'ultimo aggiornamento.
 * 2. PM2,5: come sopra.
 * 3. PM1: come sopra.
 * 4. Potenza segnale Wi-Fi: valore RSSI (in dBm) al momento del caricamento.
 * 5. Stato sensore KS0578: % di letture riuscite dall'ultimo aggiornamento.
 * 6. Stato sensore DHT22: 1 se la lettura è riuscita, 0 altrimenti.
 * 7. Temperatura: in °Celsius, rilevata al momento del caricamento.
 * 8. Umidità relativa: in %, rilevata al momento del caricamento.
 *
 * Per ulteriori dettagli implementativi, il codice è proprio qui sotto! :)
 *
 * \author Michele Nuzzolese
 * \copyright Pubblico dominio con licenza Unlicense: https://unlicense.org
 * \sa Progetto ScienziAria: https://comunitacircolare.it/progetti/scienziaria
 *
 */

#include <DHT.h>
#include <Wire.h>
#include <WiFi.h>
#include <HTTPClient.h>
#include <NetworkClientSecure.h>
#include <esp32-hal-cpu.h>

// I parametri obbligatori sono gli stessi della versione base della centralina,
// mentre i parametri opzionali sono inclusi in tutte le altre costanti globali.

// Come rete Wi-Fi può andar bene anche un hotspot del cellulare.
const char *NOME_RETE_WIFI = "hotspot";
// Password del Wi-Fi (o dell'hotspot).
const char *PASSWORD_WIFI = "scienziaria";
// Chiave API del canale ThingSpeak.
const char *WRITE_API_KEY = "4M43F4C10CH3VU01";

// I valori indicati di seguito si basano sulle istruzioni della guida e devono
// rispecchiare l'effettivo collegamento tra i componenti e la scheda ESP32:
const uint8_t PIN_SDA = 27;    // Terzo pin del KS0578 (da sinistra a destra).
const uint8_t PIN_SCL = 26;    // Quarto pin del KS0578 (da sinistra a destra).
const uint8_t PIN_DHT = 25;    // Secondo pin del DHT22 (pin DATA).
const uint8_t PIN_LED_R = 12;  // Pin rosso del LED RGB.
const uint8_t PIN_LED_G = 13;  // Pin verde del LED RGB.
const uint8_t PIN_LED_B = 14;  // Pin blu del LED RGB.

// Frequenza con cui inviare nuove rilevazioni (in millisecondi), considerando
// le limitazioni di ThingSpeak (v. https://thingspeak.com/pages/license_faq).
const uint32_t INTERVALLO_AGGIORNAMENTI_MS = 15423;

// Tempo di attesa per letture stabili dall'avvio della centralina (in secondi).
const uint8_t STABILIZZAZIONE_SENSORE_PM_S = 30;
// Frequenza con cui leggere il sensore di polveri sottili (in millisecondi).
const uint32_t INTERVALLO_MONITORAGGIO_PM_MS = 1000;
// Frequenza con cui far lampeggiare il LED (in millisecondi).
const uint32_t INTERVALLO_INTERMITTENZA_LED_MS = 500;

// Soglia oltre la quale segnalare un livello PM alto.
const uint8_t SOGLIA_PM10_ALTA = 50;
// Soglia sotto la quale segnalare un livello PM basso.
const uint8_t SOGLIA_PM10_BASSA = 25;

// Tipo di dati personalizzato per gestire i colori dell'indicatore LED.
typedef enum {
  OFF,
  RED,
  GREEN,
  BLUE,
  YELLOW,
  MAGENTA,
  CYAN,
  WHITE
} led_color_t;

// Colore del LED in base allo stato della centralina.
const led_color_t COLORE_STABILIZZAZIONE_SENSORE = CYAN,
                  COLORE_PM10_BASSO = GREEN,
                  COLORE_PM10_MEDIO = BLUE,
                  COLORE_PM10_ALTO = RED,
                  COLORE_LETTURA_FALLITA = WHITE,
                  COLORE_AGGIORNAMENTO_THINGSPEAK = MAGENTA;

// Certificato del server di ThingSpeak per la connessione sicura via HTTPS.
const char *CA_CERT =
  "-----BEGIN CERTIFICATE-----\n"
  "MIIDjjCCAnagAwIBAgIQAzrx5qcRqaC7KGSxHQn65TANBgkqhkiG9w0BAQsFADBh\n"
  "MQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3\n"
  "d3cuZGlnaWNlcnQuY29tMSAwHgYDVQQDExdEaWdpQ2VydCBHbG9iYWwgUm9vdCBH\n"
  "MjAeFw0xMzA4MDExMjAwMDBaFw0zODAxMTUxMjAwMDBaMGExCzAJBgNVBAYTAlVT\n"
  "MRUwEwYDVQQKEwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5j\n"
  "b20xIDAeBgNVBAMTF0RpZ2lDZXJ0IEdsb2JhbCBSb290IEcyMIIBIjANBgkqhkiG\n"
  "9w0BAQEFAAOCAQ8AMIIBCgKCAQEAuzfNNNx7a8myaJCtSnX/RrohCgiN9RlUyfuI\n"
  "2/Ou8jqJkTx65qsGGmvPrC3oXgkkRLpimn7Wo6h+4FR1IAWsULecYxpsMNzaHxmx\n"
  "1x7e/dfgy5SDN67sH0NO3Xss0r0upS/kqbitOtSZpLYl6ZtrAGCSYP9PIUkY92eQ\n"
  "q2EGnI/yuum06ZIya7XzV+hdG82MHauVBJVJ8zUtluNJbd134/tJS7SsVQepj5Wz\n"
  "tCO7TG1F8PapspUwtP1MVYwnSlcUfIKdzXOS0xZKBgyMUNGPHgm+F6HmIcr9g+UQ\n"
  "vIOlCsRnKPZzFBQ9RnbDhxSJITRNrw9FDKZJobq7nMWxM4MphQIDAQABo0IwQDAP\n"
  "BgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBhjAdBgNVHQ4EFgQUTiJUIBiV\n"
  "5uNu5g/6+rkS7QYXjzkwDQYJKoZIhvcNAQELBQADggEBAGBnKJRvDkhj6zHd6mcY\n"
  "1Yl9PMWLSn/pvtsrF9+wX3N3KjITOYFnQoQj8kVnNeyIv/iPsGEMNKSuIEyExtv4\n"
  "NeF22d+mQrvHRAiGfzZ0JFrabA0UWTW98kndth/Jsw1HKj2ZL7tcu7XUIOGZX1NG\n"
  "Fdtom/DzMNU+MeKNhJ7jitralj41E6Vf8PlwUHBHQRFXGU7Aj64GxJUTFy8bJZ91\n"
  "8rGOmaFvE7FBcf6IKshPECBV1/MUReXgRPTqh5Uykw7+U0b6LJ3/iyK5S9kJRaTe\n"
  "pLiaWN0bfVKfjllDiIGknibVb63dDcY3fe0Dkhvld1927jyNxF1WW6LZZm6zNTfl\n"
  "MrY=\n"
  "-----END CERTIFICATE-----";
// Limite di tempo per effettuare l'handshake SSL (in secondi).
const uint8_t TIMEOUT_HANDSHAKE_SSL_S = 4;

// Limite di tempo per stabilire la connessione con il server (in millisecondi).
const uint32_t TIMEOUT_CONNESSIONE_CLIENT_MS = 3000;
// Limite di tempo per ricevere la risposta del server (in millisecondi).
const uint32_t TIMEOUT_RISPOSTA_SERVER_MS = 3000;

// Tempo di attesa prima di ritentare dopo un aggiornamento fallito (in ms).
const uint32_t ATTESA_NUOVO_TENTATIVO_MS = 2342;

// In questa variante servono più variabili globali rispetto alla versione base,
// per gestire lo stato della centralina e l'esecuzione dei vari task (compiti).

uint16_t pm1 = 0;   // Concentrazione di PM1 rilevata dal sensore KS0578.
uint16_t pm25 = 0;  // Concentrazione di PM2,5 rilevata dal sensore KS0578.
uint16_t pm10 = 0;  // Concentrazione di PM10 rilevata dal sensore KS0578.

float temperatura = NAN;  // Temperatura rilevata dal sensore DHT22.
float umidita = NAN;      // Umidità relativa rilevata dal sensore DHT22.

DHT dht(PIN_DHT, DHT22);  // Sensore DHT22 (vedi libreria `DHT.h` di Adafruit).

NetworkClientSecure client;  // Client per la connessione sicura HTTPS.
HTTPClient httpClient;       // Client per l'invio della richiesta GET.

// Media mobile esponenziale della concentrazione PM dall'ultimo aggiornamento.
float pm10media = 0, pm25media = 0, pm1media = 0;
// Totale di letture effettuate del sensore PM dall'ultimo aggiornamento.
uint8_t lettureRiuscite = 0, lettureFallite = 0;

bool ledAcceso = false;       // Stato attuale del LED.
led_color_t coloreLed = OFF;  // Colore attuale del LED.

// Timer per il task di monitoraggio delle polveri sottili.
uint32_t millisMonitoraggio = 0;

// Timer per il task di lampeggiamento del LED.
uint32_t millisLed = 0;

// Timer per il task di aggiornamento del canale ThingSpeak.
uint32_t millisAggiornamento = 1000L * STABILIZZAZIONE_SENSORE_PM_S;
// Intervallo del prossimo task di aggiornamento del canale ThingSpeak.
uint32_t intervalloAggiornamento = INTERVALLO_AGGIORNAMENTI_MS;

// Nota: come alternativa a tutte queste costanti e variabili globali potremmo
// incapsulare la centralina in un'apposita classe (separata dalla logica della
// gestione dei task, che non riguardano l'implementazione specifica ma solo il
// comportamento definito da un'interfaccia pubblica). Sebbene questo approccio
// basato sul principio di occultamento delle informazioni (information hiding)
// sia più teoricamente corretto e ingegneristicamente robusto, introduce anche
// complessità progettuali sproporzionate per questa specifica applicazione (ma
// lo consigliamo come esercizio sulle tecniche della programmazione a oggetti).

/**
 * Legge il KS0578 (presupponendo che la comunicazione sia già stata aperta) e
 * memorizza le misurazioni di polveri sottili nelle apposite variabili globali.
 *
 * Nota: stessa funzione della versione base della centralina (con il KS0578).
 * 
 * @return `true` se la lettura ha esito positivo, `false` altrimenti.
 */
bool readPM() {

  // Assicuriamoci che il buffer I2C sia vuoto prima di ricevere nuovi dati
  while (Wire.available()) {
    Wire.read();
  }

  // Leggiamo 32 byte dal bus I2C in base al protocollo di trasporto del sensore
  if (Wire.requestFrom(0x2a, 32) != 32) {
    return false;  // risposta non conforme al protocollo di trasporto
  }

  uint8_t datiSensore[32] = { 0 };
  uint16_t checksumCalcolato = 0;

  // Prepariamo i dati da elaborare e i checksum per i controlli di integrità
  for (uint8_t i = 0; i < 32; i++) {
    datiSensore[i] = Wire.read();
    checksumCalcolato += datiSensore[i];
  }
  checksumCalcolato -= datiSensore[30] + datiSensore[31];

  // Validiamo i dati ricevuti e controlliamo l'esito della trasmissione seriale
  if (checksumCalcolato == 0) {
    return false;  // sensore non ancora disponibile (tutti i dati uguali a 0)
  }
  uint16_t checksumRicevuto = (datiSensore[30] << 8) | datiSensore[31];
  if (checksumRicevuto != checksumCalcolato) {
    return false;  // errore di trasmissione (checksum non corrispondenti)
  }

  // Elaboriamo le rilevazioni di PM e memorizziamole nelle rispettive variabili
  pm1 = (datiSensore[10] << 8) | datiSensore[11];
  pm25 = (datiSensore[12] << 8) | datiSensore[13];
  pm10 = (datiSensore[14] << 8) | datiSensore[15];

  return true;  // lettura effettuata con successo
}

/**
 * Legge il DHT22 (presupponendo che la comunicazione sia già stata aperta) e
 * memorizza le misurazioni di temperatura e umidità nelle apposite variabili.
 * 
 * @return `true` se la lettura ha esito positivo, `false` altrimenti.
 */
bool readDHT() {
  temperatura = dht.readTemperature();
  umidita = dht.readHumidity();

  if (isnan(umidita) || isnan(temperatura)) {
    return false;  // lettura fallita
  }

  return true;  // lettura effettuata con successo
}

/**
 * Imposta la spia della centralina in base al colore passato come parametro.
 *
 * @param color Il colore da impostare (oppure `OFF` per spegnere il LED).
 */
void setLed(led_color_t color) {

  if (ledAcceso && color == coloreLed) {
    return;
  }

  if (color == OFF) {
    ledAcceso = false;
    digitalWrite(PIN_LED_R, LOW);
    digitalWrite(PIN_LED_G, LOW);
    digitalWrite(PIN_LED_B, LOW);
    return;
  }

  coloreLed = color;
  ledAcceso = true;

  switch (color) {
    case GREEN:
      digitalWrite(PIN_LED_R, LOW);
      digitalWrite(PIN_LED_G, HIGH);
      digitalWrite(PIN_LED_B, LOW);
      break;
    case BLUE:
      digitalWrite(PIN_LED_R, LOW);
      digitalWrite(PIN_LED_G, LOW);
      digitalWrite(PIN_LED_B, HIGH);
      break;
    case RED:
      digitalWrite(PIN_LED_R, HIGH);
      digitalWrite(PIN_LED_G, LOW);
      digitalWrite(PIN_LED_B, LOW);
      break;
    case YELLOW:
      digitalWrite(PIN_LED_R, HIGH);
      digitalWrite(PIN_LED_G, HIGH);
      digitalWrite(PIN_LED_B, LOW);
      break;
    case MAGENTA:
      digitalWrite(PIN_LED_R, HIGH);
      digitalWrite(PIN_LED_G, LOW);
      digitalWrite(PIN_LED_B, HIGH);
      break;
    case CYAN:
      digitalWrite(PIN_LED_R, LOW);
      digitalWrite(PIN_LED_G, HIGH);
      digitalWrite(PIN_LED_B, HIGH);
      break;
    case WHITE:
      digitalWrite(PIN_LED_R, HIGH);
      digitalWrite(PIN_LED_G, HIGH);
      digitalWrite(PIN_LED_B, HIGH);
      break;
    default:
      // abbiamo già coperto esaustivamente tutte le casistiche possibili
      break;
  }
}

/**
 * Prepara la centralina alla fase di monitoraggio continuo.
 */
void setup() {

  // Impostiamo la frequenza della CPU a 80 MHz per consumare meno corrente
  setCpuFrequencyMhz(80);

  // Inizializziamo i pin del LED RGB
  pinMode(PIN_LED_R, OUTPUT);
  pinMode(PIN_LED_G, OUTPUT);
  pinMode(PIN_LED_B, OUTPUT);

  // Accendiamo il LED per indicare che il sensore è in fase di stabilizzazione
  setLed(COLORE_STABILIZZAZIONE_SENSORE);

  // Inizializziamo il Wi-Fi
  WiFi.begin(NOME_RETE_WIFI, PASSWORD_WIFI);

  // Nota: l'ESP32 implementa l'API WiFi di Arduino impostando la riconnessione
  // automatica di default, ma dai nostri test è risultato più efficace gestirla
  // con un evento (strategia consigliata anche nella documentazione dell'ESP32,
  // v. https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-guides/wifi.html#wi-fi-reconnect).

  // Disabilitiamo la riconnessione automatica
  WiFi.setAutoReconnect(false);
  // Registriamo la funzione da richiamare in caso di disconnessione dal Wi-Fi
  WiFi.onEvent(
    [](WiFiEvent_t e, WiFiEventInfo_t i) {  // gestore evento (funzione anonima)
      WiFi.reconnect();
    },
    WiFiEvent_t::ARDUINO_EVENT_WIFI_STA_DISCONNECTED);  // evento disconnessione

  // Inizializziamo la communicazione con i sensori
  Wire.begin(PIN_SDA, PIN_SCL);
  dht.begin();

  // Impostiamo il certificato del client e il timeout per l'handshake SSL/TLS
  client.setCACert(CA_CERT);
  client.setHandshakeTimeout(TIMEOUT_HANDSHAKE_SSL_S);

  // Impostiamo i timeout per la connessione al client e la risposta del server
  // (v. https://github.com/espressif/arduino-esp32/issues/1433#issuecomment-475875579)
  httpClient.setConnectTimeout(TIMEOUT_CONNESSIONE_CLIENT_MS);
  httpClient.setTimeout(TIMEOUT_RISPOSTA_SERVER_MS);

  // Aspettiamo che il sensore di polveri sottili si sia stabilizzato
  while (1000L * STABILIZZAZIONE_SENSORE_PM_S > millis()) {

    // Nel frattempo, facciamo lampeggiare il LED se il Wi-Fi è disconnesso
    if (WiFi.status() != WL_CONNECTED) {

      delay(INTERVALLO_INTERMITTENZA_LED_MS);
      setLed(OFF);
      delay(INTERVALLO_INTERMITTENZA_LED_MS);
      setLed(COLORE_STABILIZZAZIONE_SENSORE);

      // Nota: bloccare l'esecuzione per tempi ragionevolmente brevi come quello
      // di un LED che lampeggia è del tutto innocuo in questo contesto (e rende
      // anche più immediata l'implementazione, eliminando il caso limite in cui
      // il Wi-Fi dovesse riconnettersi appena prima che il LED venga riacceso).
    }
  }
}

/**
 * Monitora il PM e il LED, controlla il Wi-Fi e aggiorna il canale ThingSpeak.
 */
void loop() {

  // Controlliamo se è il momento di eseguire il task di monitoraggio
  if (millis() - millisMonitoraggio >= INTERVALLO_MONITORAGGIO_PM_MS) {
    millisMonitoraggio = millis();

    if (!readPM()) {
      lettureFallite++;
      setLed(COLORE_LETTURA_FALLITA);
    } else {
      lettureRiuscite++;

      // Qui c'è un operatore ternario innestato in un altro operatore ternario
      // (ma è meno complesso di quanto possa sembrare, vedi i commenti a lato)
      setLed(pm10 < SOGLIA_PM10_BASSA      // se questa condizione è vera
               ? COLORE_PM10_BASSO         // allora restituisce questo valore
               : (pm10 < SOGLIA_PM10_ALTA  // altrimenti: se questa è vera
                    ? COLORE_PM10_MEDIO    // allora restituisce questo valore
                    : COLORE_PM10_ALTO));  // altrimenti: restituisce questo qui

      if (lettureRiuscite == 1) {  // passo iniziale per il calcolo della media
        pm10media = pm10;
        pm25media = pm25;
        pm1media = pm1;
      } else {
        const float ALFA = 0.69;  // parametro per la media mobile esponenziale
        pm10media += ALFA * (pm10 - pm10media);
        pm25media += ALFA * (pm25 - pm25media);
        pm1media += ALFA * (pm1 - pm1media);
      }
    }  // fine controllo lettura

  }  // fine task monitoraggio

  // Segnaliamo che il WiFi è disconnesso facendo lampeggiare il LED
  if (WiFi.status() != WL_CONNECTED) {
    if (millis() - millisLed >= INTERVALLO_INTERMITTENZA_LED_MS) {
      millisLed = millis();
      setLed(ledAcceso ? OFF : coloreLed);
    }
    // Nota: bloccare l'esecuzione interferirebbe con il task di monitoraggio,
    // quindi abbiamo usato un timer per il LED e gestito l'evento per eseguire
    // la funzione di riconnessione al Wi-Fi in parallelo, in un thread a parte.
  } else {
    // Se il WiFi si riconnette mentre il LED è ancora spento, riaccendiamolo
    if (!ledAcceso) {
      setLed(coloreLed);
    }
    // Nota: questo controllo è ridondante perché svolto anche all'interno della
    // funzione, ma lo sappiamo solo perché l'abbiamo scritta noi. Se avessimo a
    // che fare con un generico metodo setter, potremmo non essere a conoscenza
    // di simili dettagli implementativi (e un controllo aggiuntivo contribuisce
    // comunque alla chiarezza del codice, a un costo computazionale irrisorio).

    // Controlliamo se è il momento di eseguire il task di aggiornamento
    if (millis() - millisAggiornamento >= intervalloAggiornamento) {

      // Elaboriamo i dati di controllo da inviare al canale ThingSpeak
      int8_t rssi = WiFi.RSSI();
      float esitoLetturePM = 100.0 * lettureRiuscite / (lettureRiuscite + lettureFallite);
      bool esitoLetturaDHT = readDHT();

      // Creiamo la richiesta nel formato definito dall'API di ThingSpeak
      String uri = "/update?api_key=" + String(WRITE_API_KEY);
      // Inviamo i dati sulle polveri sottili solo se disponibili
      if (lettureRiuscite > 0) {
        uri += "&field1=" + String(pm10media, 2)
               + "&field2=" + String(pm25media, 2)
               + "&field3=" + String(pm1media, 2);
      }
      // Inviamo sempre e comunque i dati di controllo
      uri += "&field4=" + String(rssi)
             + "&field5=" + String(esitoLetturePM, 2)
             + "&field6=" + String((int)esitoLetturaDHT);
      // Inviamo i dati su temperatura e umidità solo se disponibili
      if (esitoLetturaDHT) {
        uri += "&field7=" + String(temperatura, 2)
               + "&field8=" + String(umidita, 2);
      }

      // Predisponiamo l'invio della richiesta GET e aspettiamo la risposta
      httpClient.begin(client, "api.thingspeak.com", 443, uri, true);
      led_color_t colorePrecedente = coloreLed;
      setLed(COLORE_AGGIORNAMENTO_THINGSPEAK);
      int16_t codiceRisposta = httpClient.GET();  // funzione bloccante

      // Nota: in buone condizioni di connettività ci vuole solitamente meno di
      // mezzo secondo, mentre nei casi peggiori la richiesta andrà in timeout.
      // Il primo invio richiede fino a 2 secondi per stabilire una connessione
      // criptata. Il protocollo HTTP è più veloce ma insicuro (ed è meglio non
      // trasmettere in chiaro informazioni sensibili come la chiave API).

      // Reimpostiamo il timer del task di aggiornamento e il colore del LED
      millisAggiornamento = millis();
      setLed(colorePrecedente);

      // Aggiorniamo l'intervallo del task (in base all'esito della richiesta)
      if (codiceRisposta != HTTP_CODE_OK) {
        intervalloAggiornamento = ATTESA_NUOVO_TENTATIVO_MS;
      } else {
        intervalloAggiornamento = INTERVALLO_AGGIORNAMENTI_MS;

        // Resettiamo i contatori delle letture fino al prossimo aggiornamento
        lettureRiuscite = lettureFallite = 0;
        // (la media mobile delle rilevazioni invece si aggiorna in automatico)
      }

      // Terminiamo la richiesta appena effettuata (a prescindere dal suo esito)
      httpClient.end();

    }  // fine task aggiornamento

  }  // fine controllo wifi

}  // fine loop
